
from typing import Optional, Tuple

from pymap.exceptions import AppendFailure
from pymap.interfaces.filter import FilterInterface
from pymap.interfaces.session import AppendMessage
from sievelib.parser import Parser  # type: ignore

from .runner import SieveRunner
from .util import unquote

__all__ = ['SieveParseError', 'SieveFilter']


class SieveParseError(ValueError):
    """Parsing the sieve script failed."""
    pass


class SieveFilter(FilterInterface):
    """Filter implementation for applying sieve scripts to the appended
    message. The sieve script applies various conditions and controls to
    ultimately produce a set of actions to apply to the message.

    Args:
        script: The sieve script contents.

    """

    def __init__(self, script: bytes) -> None:
        super().__init__()
        self._script = script
        self._runner: Optional[SieveRunner] = None

    @property
    def runner(self) -> SieveRunner:
        if self._runner is None:
            p = Parser()
            if p.parse(self._script):
                self._runner = SieveRunner(p.result)
            else:
                raise SieveParseError(p.error)
        return self._runner

    async def apply(self, sender: str, recipient: str, mailbox: str,
                    append_msg: AppendMessage) \
            -> Tuple[Optional[str], AppendMessage]:
        for action in self.runner.get_actions(sender, recipient, append_msg):
            if action.name == 'keep':
                pass
            elif action.name == 'fileinto':
                mailbox = unquote(action.arguments['mailbox'])
                return mailbox, append_msg
            elif action.name == 'reject':
                msg = unquote(action.arguments['text']).encode('utf-8')
                raise AppendFailure(mailbox, msg)
            elif action.name == 'discard':
                return None, append_msg
            else:
                msg = b'Unrecognized sieve action: %b' \
                    % action.name.encode('ascii')
                raise AppendFailure(mailbox, msg) from KeyError(action.name)
        return mailbox, append_msg




from abc import abstractmethod, ABCMeta
from typing import TypeVar, Generic, Callable, Optional, NoReturn

from pkg_resources import iter_entry_points, DistributionNotFound

from .interfaces.filter import FilterInterface, FilterSetInterface

__all__ = ['FilterValueT', 'EntryPointFilterSet', 'SingleFilterSet']

#: Type variable for the value used to instantiate an
#: :class:`EntryPointFilter`.
FilterValueT = TypeVar('FilterValueT')

_FilterFactory = Callable[[FilterValueT], FilterInterface]


class EntryPointFilterSet(Generic[FilterValueT], FilterSetInterface,
                          metaclass=ABCMeta):
    """Base class for filter set implementations that use a filter
    implementation declared in the ``pymap.filter`` entry point.

    The entry points should be classes or callable factories that take a single
    value and return a filter instance.

    """

    @property
    @abstractmethod
    def entry_point(self) -> str:
        """The name of the ``pymap.filter`` entry point used to implement the
        filter.

        If the entry point is found, it must implement :class:`FilterInterface`
        for the filter value type. If the entry point is not found, no filter
        is applied to new messages.

        """
        ...

    def get_filter(self, value: FilterValueT) -> FilterInterface:
        """Find the filter implementation, and return an instance of it with
        the given filter value.

        Args:
            value: The filter value.

        Raises:
            :class:`LookupError`

        """
        filter_factory: Optional[_FilterFactory] = None
        name = self.entry_point
        for entry_point in iter_entry_points('pymap.filter', name):
            try:
                filter_factory = entry_point.load()
            except DistributionNotFound:
                pass  # optional dependencies not installed
        if filter_factory is None:
            raise LookupError(name)
        return filter_factory(value)


class SingleFilterSet(FilterSetInterface, metaclass=ABCMeta):
    """Base class for a filter set that does not use named filter
    implementations, it contains only a single active filter implementation.

    """

    async def put(self, name: str, value: FilterInterface,
                  check: bool = False) -> NoReturn:
        raise NotImplementedError()

    async def delete(self, name: str) -> NoReturn:
        raise NotImplementedError()

    async def rename(self, before_name: str, after_name: str) -> NoReturn:
        raise NotImplementedError()

    async def set_active(self, name: Optional[str]) -> NoReturn:
        raise NotImplementedError()

    async def get(self, name: str) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    async def get_active(self) -> Optional[FilterInterface]:
        ...

    async def get_all(self) -> NoReturn:
        raise NotImplementedError()



"""IMAP server with pluggable Python backends."""

import asyncio
import logging
import traceback
from argparse import ArgumentParser, Namespace, ArgumentDefaultsHelpFormatter
from typing import Any, Type, Sequence, Mapping

from pkg_resources import iter_entry_points, DistributionNotFound

from .core import __version__
from .interfaces.backend import BackendInterface, ServiceInterface

_Backends = Mapping[str, Type[BackendInterface]]
_Services = Mapping[str, Type[ServiceInterface]]


def _load_entry_points(parser: ArgumentParser, group: str) \
        -> Mapping[str, Type[Any]]:
    ret = {}
    for entry_point in iter_entry_points(group):
        try:
            cls = entry_point.load()
        except DistributionNotFound:
            pass  # optional dependencies not installed
        except ImportError:
            traceback.print_exc()
            parser.exit(1, f'Error importing: {group}:{entry_point.name}\n')
        else:
            ret[entry_point.name] = cls
    return ret


def main() -> None:
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--debug', action='store_true',
                        help='increase printed output for debugging')
    parser.add_argument('--version', action='version',
                        version='%(prog)s' + __version__)
    subparsers = parser.add_subparsers(dest='backend',
                                       help='which pymap backend to use')
    service = parser.add_argument_group('service arguments')
    service.add_argument('--no-services', action='store_true',
                         help='do not run any registered services')
    service.add_argument('--no-service', dest='skip_services', action='append',
                         metavar='NAME', help='do not run the given service')
    listener = parser.add_argument_group('server arguments')
    listener.add_argument('--port', action='store', type=int, default=1143,
                          help='the port to listen on')
    listener.add_argument('--cert', action='store', help='cert file for TLS')
    listener.add_argument('--key', action='store', help='key file for TLS')
    listener.add_argument('--insecure-login', action='store_true',
                          help='allow plaintext login without TLS')

    backends: _Backends = _load_entry_points(parser, 'pymap.backend')
    services: _Services = _load_entry_points(parser, 'pymap.service')

    for backend_cls in backends.values():
        backend_cls.add_subparser(subparsers)
    for service_cls in services.values():
        service_cls.add_arguments(parser)
    parser.set_defaults(skip_services=[])
    args = parser.parse_args()

    logging.basicConfig(level=logging.WARNING)

    if not args.backend:
        parser.error('Expected backend name')
    backend = backends[args.backend]

    if not args.no_services:
        run_services = [service for name, service in services.items()
                        if name not in args.skip_services]
    else:
        run_services = []

    try:
        return asyncio.run(run(args, backend, run_services), debug=False)
    except KeyboardInterrupt:
        pass


async def run(args: Namespace, backend_type: Type[BackendInterface],
              service_types: Sequence[Type[ServiceInterface]]) -> None:
    backend = await backend_type.init(args)
    backend.config.apply_context()
    services = [await service.init(backend) for service in service_types]

    server = await asyncio.start_server(backend, port=args.port)

    # Typeshed currently has poor stubs for AbstractServer.
    async with server:  # type: ignore
        await asyncio.gather(server.serve_forever(),  # type: ignore
                             *(service.run_forever() for service in services))


if __name__ == '__main__':
    main()
